--------------------------------------------------------------------------------
-- | 
-- Module: BitmapTools
--
-- Contains bitmap manipulation, export and import functions. Every drawing 
-- function comes in two flavors: one to draws on a bitmap directly and another 
-- one two returns a list of 'PixelChange', which allows the programmer to
-- accumulate changes before commiting them. Reasons for this are explained in 
-- documentation of module 'Bitmap'.
--
-- If error "Error in array index" is thrown, it means that you are trying to 
-- draw outside of the bitmap.
--
--------------------------------------------------------------------------------

module BitmapTools (
    -- * Drawing functions on bitmap
    drawFillRect,   -- :: Bitmap -> ColorRGB -> Rectangle -> Bitmap
    drawLine,       -- :: Bitmap -> ColorRGB -> Pixel -> Pixel -> Bitmap
    drawPoint,      -- :: Bitmap -> ColorRGB -> Pixel -> Bitmap
    drawRect,       -- :: Bitmap -> ColorRGB -> Rectangle -> Bitmap

    -- * Accumulating functions on bitmap
    drawFillRect',  -- :: ColorRGB -> Rectangle -> [PixelChange]
    drawLine',      -- :: ColorRGB -> Pixel -> Pixel -> [PixelChange]
    drawPoint',     -- :: ColorRGB -> Pixel -> [PixelChange]
    drawRect',      -- :: ColorRGB -> Rectangle -> [PixelChange]

    -- * Import functions
    importFromBMP,  -- :: String -> IO Bitmap

    -- * Export functions
    exportAsBMP     -- :: Bitmap -> String -> IO ()
    ) where

import Bitmap
import Colors
import Utils

import Data.Char (ord)
import Data.List (isSuffixOf)


--------------------------------------------------------------------------------
-- DRAWING FUNCTIONS
--------------------------------------------------------------------------------

-- | Returns a list of pixel changes, which would draw a point on a bitmap.
drawPoint' :: ColorRGB -> Pixel -> [PixelChange]
drawPoint' newColor pt = [(pt, newColor)]

-- | Draws a point on the bitmap.
drawPoint :: Bitmap -> ColorRGB -> Pixel -> Bitmap
drawPoint bitmap newColor pt = setPixels bitmap (drawPoint' newColor pt)


-- | Returns a list of pixel changes, which would draw a rectangle on a bitmap.
drawRect' :: ColorRGB -> Rectangle Int -> [PixelChange]
drawRect' rectColor (Rect (ulx, uly) (lrx, lry)) = newPoints
    where
    topBound    = [ ((x, uly), rectColor) | x <- [ulx..lrx] ]
    bottomBound = [ ((x, lry), rectColor) | x <- [ulx..lrx] ]
    leftBound   = [ ((ulx, y), rectColor) | y <- [uly..lry] ]
    rightBound  = [ ((lrx, y), rectColor) | y <- [uly..lry] ]
    newPoints   = topBound ++ bottomBound ++ leftBound ++ rightBound

-- | Draws a rectangle on a bitmap.
drawRect :: Bitmap -> ColorRGB -> Rectangle Int -> Bitmap
drawRect bitmap rectColor (Rect p1 p2) = setPixels bitmap (drawRect' rectColor (Rect p1 p2))


-- | Returns a list of pixel changes, which would draw a filled rectangle on a
-- bitmap.
drawFillRect' :: ColorRGB -> Rectangle Int -> [PixelChange]
drawFillRect' rectColor (Rect (ulx, uly) (lrx, lry)) = newPoints
    where
    newPoints = [ ((x, y), rectColor) | x <- [ulx..lrx], y <- [uly..lry] ]

-- | Draws a filled rectangle on the bitmap. 
drawFillRect :: Bitmap -> ColorRGB -> Rectangle Int -> Bitmap
drawFillRect bitmap rectColor (Rect p1 p2) = setPixels bitmap (drawFillRect' rectColor (Rect p1 p2))
    

-- Returns a list of pixel changes, which would draw a line on a bitmap. The
-- line is specified by its endpoints. 
--
-- Line is obtained using the Bresenham's line algorithm.
drawLine' :: ColorRGB -> Pixel -> Pixel -> [PixelChange]
drawLine' lineColor (p1x, p1y) (p2x, p2y) = newPoints
    where
    steepLine = (abs (p2y - p1y)) > (abs (p2x - p1x))

    (x1, y1, x2, y2) = if steepLine
                         then if p1y > p2y                 
                                then (p2y, p2x, p1y, p1x)   -- Steep, right-to-left.
                                else (p1y, p1x, p2y, p2x)   -- Steep, left-to-right.
                         else if p1x > p2x
                                then (p2x, p2y, p1x, p1y)   -- Normal, right-to-left.
                                else (p1x, p1y, p2x, p2y)   -- Normal, left-to-right.
    deltaX = x2 - x1
    deltaY = abs (y2 - y1)
    yStep  = if y1 < y2 then 1 else (-1)

    -- Line plotting function.
    drawLine' cX cY err
        | cX > x2   = []
        | err < 0   = drawLine' cX (cY + yStep) (err + deltaX)
        | otherwise = (curCoords, lineColor) : (drawLine' (cX + 1) cY (err - deltaY))
        where
        curCoords = if steepLine then (cY, cX) else (cX, cY)

    newPoints = drawLine' x1 y1 (deltaX `div` 2)

-- | Draws a line of the specified color on the bitmap. The line is specified
-- by its end points.
drawLine :: Bitmap -> ColorRGB -> Pixel -> Pixel -> Bitmap
drawLine bitmap lineColor p1 p2 = setPixels bitmap (drawLine' lineColor p1 p2)


--------------------------------------------------------------------------------
-- IMPORT FUNCTIONS
--------------------------------------------------------------------------------

-- | Creates a bitmap from a BMP file. This function can only process unindexed,
-- uncompressed 24-bit BMP files. This is a rather slow function - takes me
-- around a minute to load a 9MB image.
importFromBMP :: String -> IO Bitmap
importFromBMP filename = do rawData <- readFile filename
                            return (convertRawData rawData)
    where
    convertRawData rawData = if (isSuffixOf ".bmp" filename) && (magicNumber == "BM")
                               then if (bpp == 24) && (compression == 0) && (paletteSize == 0)
                                      then bitmap
                                      else error "Unsupported BMP file given."
                               else error "Not a BMP file given."
        where
        -- Correctness checks.
        magicNumber = slice rawData 0 1                             -- Must equal "BM"
        bpp         = rawBinToInt (reverse (slice rawData 28 29))   -- Must equal 24
        compression = rawBinToInt (reverse (slice rawData 30 33))   -- Must equal 0
        paletteSize = rawBinToInt (reverse (slice rawData 46 49))   -- Must equal 0

        -- Information about bitmap.
        dataOffset = rawBinToInt (reverse (slice rawData 10 13))
        imgWidth   = rawBinToInt (reverse (slice rawData 18 21))
        imgHeight  = rawBinToInt (reverse (slice rawData 22 25))

        (wd, ht) = (imgWidth - 1, imgHeight - 1)

        -- Converting raw data into a bitmap.
        bitmap = setPixels (makeBitmap imgWidth imgHeight) (processBMP 0 (imgHeight-1) (drop dataOffset rawData))

        offsetSize = (4 - (3 * imgWidth) `mod` 4) `mod` 4
        processBMP w h xs
            | w == imgWidth = processBMP 0 (h-1) (drop offsetSize xs)
            | h < 0         = []
            | otherwise     = ((w, h), rgbFromBinary (take 3 xs)) : (processBMP (w+1) h (drop 3 xs))

        -- Converts BMP pixel description into a ColorRGB object.
        rgbFromBinary [b,g,r] = makeRGB (ord r) (ord g) (ord b)


--------------------------------------------------------------------------------
-- EXPORT FUNCTIONS
--------------------------------------------------------------------------------

-- | Saves the given bitmap as a specified BMP file. 
exportAsBMP :: Bitmap -> String -> IO ()
exportAsBMP bitmap filename = do writeFile filename (header ++ rawData)
    where
    (width, height) = getBitmapSize bitmap
    wd = width - 1
    ht = height - 1

    -- A standard BMP header.
    rowSize  = 4 * ceiling (fromIntegral (24 * width) / 32)
    fileSize = 54 + rowSize * height

    bppField      = reverse (intToRawBin 4 24)
    fileSizeField = reverse (intToRawBin 4 fileSize)
    heightField   = reverse (intToRawBin 4 height)
    rawSizeField  = reverse (intToRawBin 4 (fileSize - 54))
    widthField    = reverse (intToRawBin 4 width)

    header = "\x42\x4D"             -- Magic number (BM).
          ++ fileSizeField          -- Size of the BMP file.
          ++ "\x00\x00"             -- Reserved.
          ++ "\x00\x00"             -- Reserved.
          ++ "\x36\x00\x00\x00"     -- Image data offset (54).
          ++ "\x28\x00\x00\x00"     -- Number of bytes in header (40).
          ++ widthField             -- Width of the bitmap.
          ++ heightField            -- Height of the bitmap.
          ++ "\x01\x00"             -- Number of color planes in use.
          ++ "\x18\x00"             -- Bytes per pixel (24).
          ++ "\x00\x00\x00\x00"     -- Compression method (none).
          ++ rawSizeField           -- Size of raw bitmap data.
          ++ "\x13\x0B\x00\x00"     -- Horizontal resolution of the image (2835).
          ++ "\x13\x0B\x00\x00"     -- Vertical resolution of the image (2835).
          ++ "\x00\x00\x00\x00"     -- Number of colors in the palette.
          ++ "\x00\x00\x00\x00"     -- Important colors (all of them).

    -- Raw image data.
    rowOffsetNum = (4 - (3 * width) `mod` 4) `mod` 4
    rowOffset    = replicate rowOffsetNum '\x00'

    rawData = concat [ concat [rawPixel x y | x <- [0..wd]] ++ rowOffset | y <- reverse [0..ht] ]

    -- Converts the color of a pixel to raw binary (little-endian order).
    rawPixel x y = blue ++ green ++ red
        where
        (RGB r g b) = getPixelColor bitmap (x,y)
        red   = intToRawBin 1 r
        green = intToRawBin 1 g
        blue  = intToRawBin 1 b


--------------------------------------------------------------------------------
-- TESTING
--------------------------------------------------------------------------------

-- | Tests the point drawing function. If test is successful, test_drawPoint.bmp 
-- should contain a 100x100 sized checker pattern.
testDrawPoint :: IO()
testDrawPoint = do
    let points = [(x,y) | x <- [0..99], y <- [0..99], ((x + y) `mod` 2) == 0]
    let bmp    = foldr (\(x,y) b -> drawPoint b white (x,y)) (makeBitmap 100 100) points
    exportAsBMP bmp "test_drawPoint.bmp"
    return ()

-- | Tests the rectangle drawing function. If test is successful, 
-- test_drawRect.bmp should contains a 3D-ish pattern from rectangles.
testDrawRect :: IO()
testDrawRect = do
    let rects = [ makeRect (x,x) (x+50, x+50) | x <- [10,12..40] ]
    let bmp   = foldr (\r b -> drawRect b white r) (makeBitmap 100 100) rects
    exportAsBMP bmp "test_drawRect.bmp"
    return ()

-- | Tests the filled rectangle drawing function. If the test is successful, file
-- test_drawFillRect.bmp should contain a nice gradient pattern from squares.
testDrawFillRect :: IO()
testDrawFillRect = do
    let rects = [ (makeRect (x,x) (x+50,x+50), makeRGB (x*6) (x*6) (x*6)) | x <- [10,12..40] ]
    let bmp   = foldr (\(r,c) b -> drawFillRect b c r) (makeBitmap 100 100) rects
    exportAsBMP bmp "test_drawFillRect.bmp"
    return ()

-- | Tests the line drawing function. On success, file test_drawLine.bmp should
-- contain a star.
testDrawLine :: IO ()
testDrawLine = do
    let lines = concat [ [((50,50), (10,x)), ((50,50), (90,x)), 
                          ((50,50), (x,10)), ((50,50), (x,90))] | x <- [10,20..90] ]
    let bmp   = foldr (\(x,y) b -> drawLine b white x y) (makeBitmap 100 100) lines
    exportAsBMP bmp "test_drawLine.bmp"
    return ()

