--------------------------------------------------------------------------------
-- | 
-- Module: Material
--
-- Defines various textures and materials, which can be applied to primitives
-- to define their appearance.
-- 
--------------------------------------------------------------------------------

module Material (
    -- * Texture metadata
    TextureMeta,

    defaultTexMeta, -- :: TextureMeta
    makeTexMeta,    -- :: (Double, Double) -> (Double, Double) -> TextureMeta

    texOffset,      -- :: TextureMeta -> (Double, Double)
    texScale,       -- :: TextureMeta -> (Double, Double)

    -- * Textures
    Texture,

    noTexture,      -- :: Texture
    makeCheckerTex, -- :: ColorRGB -> ColorRGB -> TextureMeta -> Texture
    makeBitmapTex,  -- :: Bitmap -> TextureMeta -> Texture

    texelColor,     -- :: Texture -> Double -> Double -> ColorVector
    textureSize,    -- :: Texture -> (Int, Int)

    -- * Material
    Material,

    makeMaterial,   -- :: ColorRGB -> Texture -> Double -> Double -> Int -> Material
    
    noMaterial,         -- :: Material
    solidMaterial,      -- :: ColorRGB -> Material
    textureMaterial,    -- :: Texture -> Material

    matColor,       -- :: Material -> ColorVector
    matDiffusion,   -- :: Material -> Double
    matReflection,  -- :: Material -> Double
    matShininess,   -- :: Material -> Int
    matTexture      -- :: Material -> Texture

    ) where

import Bitmap
import Colors
import Ray
import Utils


--------------------------------------------------------------------------------
-- TEXTURE METADATA
--------------------------------------------------------------------------------

-- | The metadata of a texture. Includes the amounts by which the texture
-- should be offset (the first pair) and scaled (the second pair).
--
-- I believe that this should be replaced with a 2x2 transformation matrix.
data TextureMeta = TexMeta (Double, Double) (Double, Double)
                   deriving (Eq, Read, Show)


-- | Creates the texture metadata object. Arguments, in order, are x and y axis
-- offsets and x and y axis scale factors.
makeTexMeta :: (Double, Double) -> (Double, Double) -> TextureMeta
makeTexMeta (offX, offY) (scX, scY)
    | scX <= 0  = error "Texture scale factor cannot be non-positive."
    | scY <= 0  = error "Texture scale factor cannot be non-positive."
    | otherwise = TexMeta (offX, offY) (1.0 / scX, 1.0 / scY)

-- | Returns default metadata for a texture (no offsetting or scaling).
defaultTexMeta :: TextureMeta
defaultTexMeta = makeTexMeta (0.0, 0.0) (1.0, 1.0)


-- | Returns the offsets of the texture, as a pair.
texOffset :: TextureMeta -> (Double, Double)
texOffset (TexMeta off _) = off

-- | Returns the scale factors of the texture, as a pair.
texScale :: TextureMeta -> (Double, Double)
texScale (TexMeta _ sc) = sc


--------------------------------------------------------------------------------
-- TEXTURE TYPE
--------------------------------------------------------------------------------

-- | The texture type. Different constructors allow to create a texture from
-- different sources.
data Texture = 
    -- | No texture.
    NoTex
    -- | A checker pattern with 2x2 size.
  | CheckerTex ColorVector ColorVector TextureMeta
    -- | A 2D raster texture.
  | BitmapTex (Int,Int) Bitmap TextureMeta
    deriving (Eq, Read, Show)


-- | No texture constructor.
noTexture :: Texture
noTexture = NoTex

-- | Creates a checker pattern texture. 
makeCheckerTex :: ColorRGB -> ColorRGB -> TextureMeta -> Texture
makeCheckerTex col1 col2 meta = CheckerTex (rgb2vector col1) (rgb2vector col2) meta

-- | Creates a texture from a bitmap.
makeBitmapTex :: Bitmap -> TextureMeta -> Texture
makeBitmapTex bitmap meta = BitmapTex (getBitmapSize bitmap) bitmap meta


-- | Returns the size of the texture.
textureSize :: Texture -> (Int, Int)
textureSize NoTex                = (1, 1)
textureSize (CheckerTex _ _ _)   = (2, 2)
textureSize (BitmapTex size _ _) = size

-- | Returns the color of the texture at the specified point (u, v).
texelColor :: Texture -> Double -> Double -> ColorVector
texelColor NoTex                  u v = (1.0, 1.0, 1.0)
texelColor tex@(CheckerTex _ _ _) u v = checkerTexColor tex u v
texelColor tex@(BitmapTex _ _ _)  u v = bitmapTexColor tex u v


--------------------------------------------------------------------------------
-- TEXEL COLOR CALCULATIONS
--------------------------------------------------------------------------------

-- | Returns the color of the specified bitmap texture at the given point.
-- Also applies bilinear filtering to the texture.
bitmapTexColor (BitmapTex (w, h) bitmap meta) u v = texColor
    where
    -- Texture metadata
    (offX, offY) = texOffset meta
    (scX,  scY)  = texScale meta

    -- Modified texture coordinates
    newU = scX * u + offX
    newV = scY * v + offY

    -- Fraction part of the coordinates
    fU = newU - (fromIntegral (floor newU))
    fV = newV - (fromIntegral (floor newV))

    -- Weight of each of the four sub-pixels
    w1 = (1 - fU) * (1 - fV)
    w2 = fU * (1 - fV)
    w3 = (1 - fU) * fV
    w4 = fU * fV

    -- Coordinates of each of the four sub-pixels
    u1 = (floor newU)     `mod` w
    v1 = (floor newV)     `mod` h
    u2 = (floor newU + 1) `mod` w
    v2 = (floor newV + 1) `mod` h

    -- Colors of each of the four sub-pixels.
    c1 = rgb2vector (getPixelColor bitmap (u1, v1))
    c2 = rgb2vector (getPixelColor bitmap (u2, v1))
    c3 = rgb2vector (getPixelColor bitmap (u1, v2))
    c4 = rgb2vector (getPixelColor bitmap (u2, v2))

    -- Weighted sum of colors of sub-pixels.
    texColor = addListV [ multScV w1 c1, multScV w2 c2, multScV w3 c3, multScV w4 c4 ]


-- | Returns the color of the specified checker texture at the given location.
checkerTexColor (CheckerTex col1 col2 meta) u v = 
    if ((u' + v') `mod` 2 == 0) then col1 else col2
    where
    -- Texture metadata
    (offX, offY) = texOffset meta
    (scX,  scY)  = texScale meta

    -- Modified texture coordinates
    u' = floor (u * scX + offX)
    v' = floor (v * scY + offY)


--------------------------------------------------------------------------------
-- MATERIAL TYPE
--------------------------------------------------------------------------------

-- | Definition of material type. Arguments define the color of the material,
-- its texture and its diffusion, reflection and shininess ratings.
data Material = Mat Vector Texture Double Double Int
                deriving(Eq, Read, Show)


-- | Creates a material from given color, texture and diffusion, reflection 
-- and shininess ratings.
makeMaterial :: ColorRGB -> Texture -> Double -> Double -> Int -> Material
makeMaterial col tex diff refl shine
    | (diff < 0.0) || (diff > 1.0) = error "Invalid diffusion rating."
    | (refl < 0.0) || (refl > 1.0) = error "Invalid reflection rating."
    | (shine < 0)                  = error "Invalid shininess rating."
    | otherwise                    = Mat (rgb2vector col) tex diff refl shine


-- | Creates a reflective material with specified color.
noMaterial :: Material
noMaterial = makeMaterial white noTexture 1.0 0.0 20

-- | Creates a solid material with specified color.
solidMaterial :: ColorRGB -> Material
solidMaterial color = makeMaterial color noTexture 1.0 0.0 20

-- | Creates a solid material with a texture.
textureMaterial :: Texture -> Material
textureMaterial tex = makeMaterial white tex 1.0 0.0 20


-- | Returns the color of the material.
matColor :: Material -> ColorVector
matColor (Mat col _ _ _ _) = col

-- | Returns the diffusion rating of the material.
matDiffusion :: Material -> Double
matDiffusion (Mat _ _ diff _ _) = diff

-- | Returns the reflection rating of the material.
matReflection :: Material -> Double
matReflection (Mat _ _ _ refl _) = refl

-- | Returns the shininess rating of the material.
matShininess :: Material -> Int
matShininess (Mat _ _ _ _ shine) = shine

-- | Returns the texture of the material.
matTexture :: Material -> Texture
matTexture (Mat _ tex _ _ _) = tex


