--------------------------------------------------------------------------------
-- | 
-- Module: Primitives
--
-- Defines objects to be rendered and their properties. Currently has a sphere,
-- a plane, an axis-aligned box, a point light and portals defined.
-- 
--------------------------------------------------------------------------------

module Primitives (
    -- * Types
    Intersection(..),
    Primitive,

    -- * Creation of primitives
    makeBox,        -- :: Vector -> Vector -> Material -> Primitive
    makePlane,      -- :: Vector -> Double -> Material -> Primitive
    makePointLight, -- :: Vector -> ColorRGB -> Primitive
    makePortal,     -- :: Primitive -> Vector -> Primitive
    makeSphere,     -- :: Vector -> Double -> Material -> Primitive

    -- * Information about primitives
    getNormal,      -- :: Primitive -> Vector -> Vector
    intersects,     -- :: Primitive -> Ray -> Intersection

    isLight,        -- :: Primitive -> Bool
    getLightSource, -- :: Primitive -> Vector

    isPortal,       -- :: Primitive -> Bool
    getPortalDest,  -- :: Primitive -> Ray -> Ray

    -- * Material properties
    getDiffuse,         -- :: Primitive -> Double
    getMaterialColor,   -- :: Primitive -> ColorVector
    getReflection,      -- :: Primitive -> Double
    getShininess,       -- :: Primitive -> Int
    getSpecular,        -- :: Primitive -> Double
    getTextureColor     -- :: Primitive -> Vector -> ColorVector
    ) where

import Colors
import Material
import Ray
import Utils


--------------------------------------------------------------------------------
-- INTERSECTION TYPE
--------------------------------------------------------------------------------

-- | Enumeration, defining the type of ray-primitive intersection. Constructors
-- 'Outside' and 'Inside' take additional Double parameter, which denotes
-- distance to the intersection.
data Intersection = Miss                -- ^ No intersection.
                  | Outside Double      -- ^ Intersection with outside.
                  | Inside Double       -- ^ Intersection from inside.
                  deriving(Eq, Read, Show)


--------------------------------------------------------------------------------
-- SCENE PRIMITIVES
--------------------------------------------------------------------------------

-- | Defines all primitives, which can be rendered.
data Primitive =
    -- | The axis-aligned box. It is defined in terms of coordinates of its
    -- opposite corners.
    Box Vector Vector Material
    -- | The plane. It is defined by its normal and distance to the origin. 
  | Plane Vector Double Material
    -- | Point light source. Arguments specify its location, radius and material.
  | PointLight Vector Double Material
    -- | The sphere. Its defined by the location of its centre and radius.
  | Sphere Vector Double Material

    -- | The portal. This primitive is special, as it takes another primitive as
    -- its argument and gives it the properties of a portal. Any incoming rays
    -- are moved to another location (possibly with a changed direction).
  | Portal Primitive Vector
  deriving(Eq, Read, Show)


-- | Returns the normal of some primitive at the given point. It is assumed
-- that the point given is on the primitive.
getNormal :: Primitive -> Vector -> Vector
getNormal (Box c1 c2 _)        pt = boxNormal c1 c2 pt
getNormal (Plane nrm _ _)      _  = nrm
getNormal (PointLight ctr _ _) pt = normalizeV (subV pt ctr)
getNormal (Sphere ctr rad _)   pt = normalizeV (subV pt ctr)
getNormal (Portal prim _)      pt = getNormal prim pt

-- | Returns, whether some ray intersects the given primitive.
intersects :: Primitive -> Ray -> Intersection
intersects (Box c1 c2 _)          ray = rayIntersectsBox c1 c2 ray
intersects (Plane nrm dist _)     ray = rayIntersectsPlane nrm dist ray
intersects (PointLight ctr rad _) ray = rayIntersectsSphere ctr rad ray
intersects (Sphere ctr rad _)     ray = rayIntersectsSphere ctr rad ray
intersects (Portal prim _)        ray = intersects prim ray


-- | Returns True, if some primitive is a light source.
isLight :: Primitive -> Bool
isLight (PointLight _ _ _) = True
isLight _                  = False

-- | Returns the location of the light source. Return value is likely to
-- change, especially if box lights will be added.
getLightSource :: Primitive -> Vector
getLightSource (PointLight ctr _ _) = ctr
getLightSource _                    = error "Not a light source."


-- | Returns True, if some primitive is a portal.
isPortal :: Primitive -> Bool
isPortal (Portal _ _) = True
isPortal _            = False

-- | Returns the ray, as it went through a portal. The incoming ray should
-- have its origin at the intersection point and its direction should remain
-- unchanged. 
--
-- Currently it does not perform any ray rotation - only the origin is
-- affected.
getPortalDest :: Primitive -> Ray -> Ray
getPortalDest (Portal _ off) (intrPt, dir) = makeRay (subV intrPt off) dir
getPortalDest _              _             = error "Not a portal."


-- | Returns the material of some primitive.
getMaterial :: Primitive -> Material
getMaterial (Box        _ _ mat) = mat
getMaterial (Plane      _ _ mat) = mat
getMaterial (PointLight _ _ mat) = mat
getMaterial (Sphere     _ _ mat) = mat
getMaterial (Portal     prim _)  = getMaterial prim


-- | Returns the color of the primitive's material.
getMaterialColor :: Primitive -> Vector
getMaterialColor prim = matColor (getMaterial prim)

-- | Returns the diffuse rating of a primitive.
getDiffuse :: Primitive -> Double
getDiffuse prim       = matDiffusion (getMaterial prim)

-- | Returns the reflection rating of a primitive.
getReflection :: Primitive -> Double
getReflection prim    = matReflection (getMaterial prim)

-- | Returns the shininess rating of a primitive.
getShininess :: Primitive -> Int
getShininess prim     = matShininess (getMaterial prim)

-- | Returns the specular reflection rating of a primitive.
getSpecular :: Primitive -> Double
getSpecular prim      = 1.0 - matDiffusion (getMaterial prim)



-- | Returns the color of primitive's texture at specified point. It is 
-- assumed that the said point lays on the primitive.
getTextureColor :: Primitive -> Vector -> ColorVector
getTextureColor prim@(Sphere _ _ _) pt = sphereTexColor prim pt
getTextureColor prim@(Box _ _ _)    pt = boxTexColor prim pt
getTextureColor prim@(Plane _ _ _)  pt = planeTexColor prim pt
getTextureColor _                   _  = (1.0, 1.0, 1.0)



--------------------------------------------------------------------------------
-- PRIMITIVE CONSTRUCTION FUNCTIONS
--------------------------------------------------------------------------------

-- | Creates a box. The arguments are the coordinates of opposite corners of
-- the box and its material.
makeBox :: Vector -> Vector -> Material -> Primitive
makeBox (x1,y1,z1) (x2,y2,z2) mat = Box (a1,b1,c1) (a2,b2,c2) mat
    where
    (a1, a2) = minMax x1 x2
    (b1, b2) = minMax y1 y2
    (c1, c2) = minMax z1 z2

-- | Creates a plane. The arguments, in order, are normal of the plane, its 
-- distance from the origin and the material of the plane.
makePlane :: Vector -> Double -> Material -> Primitive
makePlane normal dist mat = Plane (normalizeV normal) dist mat

-- | Creates a point light source. The arguments, in order, are centre of the
-- light, its radius and color. Radius of the light does not affect its effects,
-- only the appearance.
makePointLight :: Vector -> Double -> ColorRGB -> Primitive
makePointLight centre radius color 
    | radius <= 0.0 = error "Sphere radius cannot be non-positive."
    | otherwise     = PointLight centre radius (solidMaterial color)

-- | Creates a sphere. The arguments, in order, are centre of the sphere, its
-- radius and material.
makeSphere :: Vector -> Double -> Material -> Primitive
makeSphere centre radius mat
    | radius <= 0.0 = error "Sphere radius cannot be negative."
    | otherwise     = Sphere centre radius mat

-- | Creates a portal. This function takes a primitive, which will be
-- transformed into a portal, and ray's offset.
makePortal :: Primitive -> Vector -> Primitive
makePortal prim off
    | isLight prim  = error "A light source cannot be turned into a portal."
    | isPortal prim = error "Another portal cannot be turned into a portal."
    | otherwise     = Portal prim off


--------------------------------------------------------------------------------
-- INTERSECTION CHECKS
--------------------------------------------------------------------------------

-- | Check whether a ray intersects some box. The box is defined with the
-- coordinates of its opposite corners. 
--
-- Intersections are found by checking whether a ray intersects "slabs" of a
-- box. Unless the ray is parallel to the slab, there are always two
-- intersections. By calculating all of these intersections and comparing them,
-- it is possible to find whether a ray intersects a box and where.
rayIntersectsBox :: Vector -> Vector -> Ray -> Intersection
rayIntersectsBox (a1,a2,a3) (b1,b2,b3) ((o1,o2,o3), (d1,d2,d3)) =
    if isParallel || (nearDist > farDist) || (farDist < 0)
      then Miss
      else if nearDist < 0
             then Inside farDist
             else Outside nearDist
    where
    (xMin, xMax) = slabIntrs a1 b1 o1 d1
    (yMin, yMax) = slabIntrs a2 b2 o2 d2
    (zMin, zMax) = slabIntrs a3 b3 o3 d3

    nearDist = xMin `max` yMin `max` zMin
    farDist  = xMax `min` yMax `min` zMax

    -- This variable is true, when the ray is parallel to the box and does not
    -- intersect it.
    isParallel = not $! ( ((d1 /= 0.0) || ((o1 >= a1) && (o1 <= b1))) &&
                          ((d2 /= 0.0) || ((o2 >= a2) && (o2 <= b2))) &&
                          ((d3 /= 0.0) || ((o3 >= a3) && (o3 <= b3))) )

    -- Returns the distances of both ray/slab intersections. All coordinates
    -- must be on one axis.
    slabIntrs cLow cHigh cOrig cDir = (dNear, dFar)
        where
        (dNear, dFar) = minMax ((cLow - cOrig) / cDir) ((cHigh - cOrig) / cDir)


-- | Returns, whether the given ray intersects some plane, which is defined
-- by its normal and distance to the origin.
rayIntersectsPlane :: Vector -> Double -> Ray -> Intersection
rayIntersectsPlane normal dist (orig, dir) = 
    if (den == 0) || (iDist <= 0)
      then Miss 
      else Outside iDist
    where
    den   = dotProduct dir normal
    iDist = ((-dist) - (dotProduct orig normal)) / den


-- | Returns, whether the given ray intersects some sphere, which is defined
-- by its centre coordinates and radius.
rayIntersectsSphere :: Vector -> Double -> Ray -> Intersection
rayIntersectsSphere (c1,c2,c3) radius ((o1,o2,o3), (d1,d2,d3)) = result
    where
    -- Coefficients of the quadratic equation.
    b   = 2 * (d1*(o1-c1) + d2*(o2-c2) + d3*(o3-c3))
    c   = (o1-c1)^2 + (o2-c2)^2 + (o3-c3)^2 - radius^2
    dis = b^2 - 4*c

    -- Solutions of the quadratic equation.
    sol1 = (((-b) - (sqrt dis)) / 2.0)
    sol2 = (((-b) + (sqrt dis)) / 2.0)

    result = if dis > 0
               then case (sol1 <= 0, sol2 <= 0) of
                        (True,  True ) -> Miss          -- Both intersections behind.
                        (True,  False) -> Inside sol2   -- One behind, one in front.
                        (False, False) -> Outside sol1  -- Both intersections in front.
               else Miss


--------------------------------------------------------------------------------
-- NORMAL CALCULATIONS
--------------------------------------------------------------------------------

-- | Returns the normal to a box at the specified point. 
boxNormal :: Vector -> Vector -> Vector -> Vector
boxNormal (a1,a2,a3) (b1,b2,b3) (p1,p2,p3) = nrmMatch
    where
    -- Coordinate pairs to be matched.
    matchPairs = [(p1, a1), (p1, b1), (p2, a2), (p2, b2), (p3, a3), (p3, b3)]
    -- The corresponding normals.
    normals    = [(-1, 0, 0), ( 1, 0, 0), ( 0,-1, 0), ( 0, 1, 0), ( 0, 0,-1), ( 0, 0, 1)]

    -- Finding all matches.
    pairEquals = map (\(a,b) -> (abs (a-b)) < epsilon) matchPairs
    -- Filtering matches and taking the first one.
    nrmMatch   = (snd . head) (filter fst (zip pairEquals normals))


--------------------------------------------------------------------------------
-- TEXTURE COLOR CALCULATIONS
--------------------------------------------------------------------------------

-- | Returns the color of a texture of a box at the specified point.
boxTexColor (Box (a1,b1,c1) (a2,b2,c2) mat) (x, y, z) = texColor
    where
    -- Needs more testing with proper textures.
    (u, v)   =    if (floatsEqual a1 x) then ((z-c1) / (c2-c1), (y-b1) / (b1-b2))
             else if (floatsEqual a2 x) then ((z-c1) / (c2-c1), (y-b1) / (b1-b2))
             else if (floatsEqual b1 y) then ((x-a1) / (a2-a1), (z-c1) / (c2-c1))
             else if (floatsEqual b2 y) then ((x-a1) / (a2-a1), (z-c1) / (c2-c1))
             else if (floatsEqual c1 z) then ((x-a1) / (a2-a1), (y-b1) / (b1-b2))
             else                            ((x-a1) / (a2-a1), (y-b1) / (b1-b2))

    (tW, tH) = textureSize (matTexture mat)
    texColor = texelColor (matTexture mat) (u * (fromIntegral tW)) (v * (fromIntegral tH))

    floatsEqual x y = abs (x - y) < epsilon


-- | Returns the color of a texture of a plane at the specified point.
planeTexColor (Plane nrm _ mat) pt = texColor
    where
    (x, y, z) = pt

    xAxis = normalizeV (crossVector (0, 1, 0) nrm)
    yAxis = normalizeV (crossVector xAxis (multScV (-1) nrm))
    
    (u, v) = if nrm == (0, 1, 0)
             then (x, z)
             else (dotProduct xAxis pt, dotProduct yAxis pt)

    (tW, tH) = textureSize (matTexture mat)
    texColor = texelColor (matTexture mat) (u * (fromIntegral tW) / 16) (v * (fromIntegral tH) / 16)


-- | Returns the color of a texture of a sphere at the specified point.
sphereTexColor (Sphere ctr rad mat) pt = texColor
    where
    (x, y, z) = multScV (1.0 / rad) (subV pt ctr)
    u = 1.0 - ((atan2 x z) / (2 * pi) + 0.5)
    v = 1.0 - ((asin y) / pi + 0.5)

    (tW, tH) = textureSize (matTexture mat)
    texColor = texelColor (matTexture mat) (u * (fromIntegral tW)) (v * (fromIntegral tH))
