--------------------------------------------------------------------------------
-- | 
-- Module: Ray
--
-- Defines the vector (three elements) and ray types, as well as various 
-- operations on them. 
--
-- It also contains a limited matrix implementation (4x4 size only, very 
-- specific operations on it). Originally it was needed to implement the free
-- camera, but I have found an easier way to create it without using 
-- transformation matrices. Still, they may be of use, so I am keeping them.
--
--------------------------------------------------------------------------------

module Ray (
    -- * Vectors
    Vector,
    ColorVector,

    addV,           -- :: Vector -> Vector -> Vector
    subV,           -- :: Vector -> Vector -> Vector
    multV,          -- :: Vector -> Vector -> Vector
    multScV,        -- :: Double -> Vector -> Vector
    lengthV,        -- :: Vector -> Double
    lengthSqrV,     -- :: Vector -> Double
    dotProduct,     -- :: Vector -> Vector -> Double
    crossProduct,   -- :: Vector -> Vector -> Double
    crossVector,    -- :: Vector -> Vector -> Vector
    normalizeV,     -- :: Vector -> Vector
    negateV,        -- :: Vector -> Vector

    addListV,       -- :: [Vector] -> Vector

    vector2rgb,     -- :: ColorVector -> ColorRGB
    rgb2vector,     -- :: ColorRGB -> ColorVector

    -- * Rays
    Ray,

    makeRay,        -- :: Vector -> Vector -> Ray

    -- * Matrices
    Matrix,

    makeMatrix,         -- :: [[Double]] -> Matrix
    getMatrixCell,      -- :: Matrix -> (Int, Int) -> Double
    setMatrixCell,      -- :: Matrix -> (Int, Int) -> Double -> Matrix

    identityMatrix,     -- :: Matrix
    rotationMatrix,     -- :: Double -> Double -> Double -> Matrix

    appendVector,       -- :: Matrix -> Vector -> Matrix
    invertMatrix,       -- :: Matrix -> Matrix
    multM,              -- :: Matrix -> Matrix -> Matrix
    multVectorM         -- :: Matrix -> Vector -> Vector
    ) where

import Colors
import Utils

import Data.List (transpose)
import Test.QuickCheck


--------------------------------------------------------------------------------
-- VECTOR TYPE AND OPERATIONS ON IT
--------------------------------------------------------------------------------

-- | Defines a vector with three components.
type Vector = (Double, Double, Double)

-- | A vector, which represents a color.
type ColorVector = Vector


-- | Adds two vectors together.
addV :: Vector -> Vector -> Vector
addV (a1,b1,c1) (a2,b2,c2) = (a1+a2, b1+b2, c1+c2)

-- | Subtracts one vector from another.
subV :: Vector -> Vector -> Vector
subV (a1,b1,c1) (a2,b2,c2) = (a1-a2, b1-b2, c1-c2)

-- | Multiplies two vectors together.
multV :: Vector -> Vector -> Vector
multV (a1,b1,c1) (a2,b2,c2) = (a1*a2, b1*b2, c1*c2)

-- | Multiplies a vector by a scalar.
multScV :: Double -> Vector -> Vector
multScV r (a,b,c) = (a*r, b*r, c*r)

-- | Returns the length of a vector.
lengthV :: Vector -> Double
lengthV (a,b,c) = sqrt (a*a + b*b + c*c)

-- | Returns the length of a vector squared.
lengthSqrV :: Vector -> Double
lengthSqrV (a,b,c) = a*a + b*b + c*c

-- | Returns the dot product of two vectors.
dotProduct :: Vector -> Vector -> Double
dotProduct (a1,b1,c1) (a2,b2,c2) = a1*a2 + b1*b2 + c1*c2

-- | Returns the cross product of two vectors.
crossProduct :: Vector -> Vector -> Double
crossProduct (a1,b1,c1) (a2,b2,c2) =  b1*c2 - c1*b2 + c1*a2 - a1*c2 + a1*b2 - b1*a2

-- | Returns a vector, which is perpendicular to the plane, defined by the
-- given vectors (the proper cross product).
crossVector :: Vector -> Vector -> Vector
crossVector (a1,b1,c1) (a2,b2,c2) = (b1*c2 - c1*b2, c1*a2 - c2*a1, a1*b2 - b1*a2)

-- | Normalizes a vector.
normalizeV :: Vector -> Vector
normalizeV v = multScV (1 / (lengthV v)) v

-- | Negates a vector.
negateV :: Vector -> Vector
negateV (a,b,c) = (-a,-b,-c)


-- | Adds a list of vectors together.
addListV :: [Vector] -> Vector
addListV vs = foldr addV (0,0,0) vs


-- | Converts a vector, whose elements are between 0.0 and 1.0, into an RGB
-- color.
vector2rgb :: ColorVector -> ColorRGB
vector2rgb (a,b,c) = makeRGB red green blue
    where
    red   = floor (restrict 0 255 (a * 255))
    green = floor (restrict 0 255 (b * 255))
    blue  = floor (restrict 0 255 (c * 255))

-- | Converts a RGB color into a vector.
rgb2vector :: ColorRGB -> ColorVector
rgb2vector (RGB r g b) = (a1, b1, c1)
    where
    a1 = (fromIntegral r) / 255
    b1 = (fromIntegral g) / 255
    c1 = (fromIntegral b) / 255


--------------------------------------------------------------------------------
-- RAY TYPE AND OPERATIONS ON IT
--------------------------------------------------------------------------------

-- | The ray type. First vector denotes the origin of the ray, second one
-- shows its direction.
type Ray = (Vector, Vector)

-- | Creates a ray from two vectors.
makeRay :: Vector -> Vector -> Ray
makeRay orig dir = (orig, normalizeV dir)


--------------------------------------------------------------------------------
-- MATRICES AND OPERATIONS ON THEM
--------------------------------------------------------------------------------

-- | A limited matrix data type. It can only have the size of 4x4 and very
-- specific operations (to implement rotation matrices, basically).
data Matrix = Mtx [[Double]]
              deriving (Eq, Read, Show)

-- | Creates a matrix from a list of lists of doubles.
makeMatrix :: [[Double]] -> Matrix
makeMatrix (a:b:c:d:_)
    | (length a < 4) = error "One of the matrix columns is too small."
    | (length b < 4) = error "One of the matrix columns is too small."
    | (length c < 4) = error "One of the matrix columns is too small."
    | (length d < 4) = error "One of the matrix columns is too small."
    | otherwise      = Mtx ([take 4 a] ++ [take 4 b] ++ [take 4 c] ++ [take 4 d])


-- | Returns the specified cell from a matrix.
getMatrixCell :: Matrix -> (Int, Int) -> Double
getMatrixCell (Mtx xss) (x, y)
    | (x < 0) || (x >= 4) = error "Matrix index out of bounds."
    | (y < 0) || (y >= 4) = error "Matrix index out of bounds."
    | otherwise           = (xss !! x) !! y

-- | Sets the specified value of the matrix to a new value.
setMatrixCell :: Matrix -> (Int, Int) -> Double -> Matrix
setMatrixCell (Mtx xss) (x, y) newVal
    | (x < 0) || (x >= 4) = error "Matrix index out of bounds."
    | (y < 0) || (y >= 4) = error "Matrix index out of bounds."
    | otherwise           = Mtx newData
    where
    newRow  = [(take y (xss !! x)) ++ [newVal] ++ (drop (y+1) (xss !! x))] 
    newData = (take x xss) ++ newRow ++ (drop (x+1) xss)


-- | Returns a 4x4 identity matrix.
identityMatrix :: Matrix
identityMatrix = Mtx [[(if x == y then 1 else 0) | y <- [0..3]] | x <- [0..3]]

-- | Returns a rotation matrix, which rotates by specified angles. Angles are
-- specified with degrees.
rotationMatrix :: Double -> Double -> Double -> Matrix
rotationMatrix xAngle yAngle zAngle = rotMtx
    where
    -- Radians to rotate.
    xRad = degToRad xAngle
    yRad = degToRad yAngle
    zRad = degToRad zAngle

    -- X axis rotation matrix.
    xSin = sin xRad
    xCos = cos xRad
    xRot = makeMatrix [[1,0,0,0], [0,xCos,xSin,0], [0,-xSin,xCos,0], [0,0,0,1]]

    -- Y axis rotation matrix.
    ySin = sin yRad
    yCos = cos yRad
    yRot = makeMatrix [[yCos,0,ySin,0], [0,1,0,0], [-ySin,0,yCos,0], [0,0,0,1]]

    -- Z axis rotation matrix.
    zSin = sin zRad
    zCos = cos zRad
    zRot = makeMatrix [[zCos,zSin,0,0], [-zSin,zCos,0,0], [0,0,1,0], [0,0,0,1]]

    -- The combined rotation matrix.
    rotMtx = multM xRot (multM yRot zRot)

    -- A function, which converts degrees to radians.
    degToRad x = x / 57.29577951308232


-- | Adds a vector to the fourth column to the matrix.
appendVector :: Matrix -> Vector -> Matrix
appendVector (Mtx (a:b:c:d:_)) (x,y,z) = Mtx (newA:newB:newC:d:[])
    where
    newA = (take 3 a) ++ [x]
    newB = (take 3 b) ++ [y]
    newC = (take 3 c) ++ [z]

-- | Inverts the given matrix.
invertMatrix :: Matrix -> Matrix
invertMatrix (Mtx xss) = newMtx
    where
    [[a1,a2,a3,tX],[b1,b2,b3,tY],[c1,c2,c3,tZ],_] = xss
    newTX = a1 * tX + b1 * tY + c1 * tZ
    newTY = a2 * tX + b2 * tY + c2 * tZ
    newTZ = a3 * tX + b3 * tY + c3 * tZ

    yss    = [[a1,b1,c1,newTX],[a2,b2,c2,newTY],[a3,b3,c3,newTZ],[0,0,0,1]]
    newMtx = makeMatrix yss

-- | Multiplies two matrices together.
multM :: Matrix -> Matrix -> Matrix
multM (Mtx xss) (Mtx yss) = Mtx [[multRows x y | y <- yss'] | x <- xss]
    where
    yss' = transpose yss
    multRows x y = sum [a * b | (a, b) <- zip x y]

-- | Multiplies a vector by a matrix.
multVectorM :: Matrix -> Vector -> Vector
multVectorM (Mtx xss) (a,b,c) = (x,y,z)
    where
    -- Gross
    x = a * ((xss !! 0) !! 0) + b * ((xss !! 0) !! 1) + c * ((xss !! 0) !! 2)
    y = a * ((xss !! 1) !! 0) + b * ((xss !! 1) !! 1) + c * ((xss !! 1) !! 2)
    z = a * ((xss !! 2) !! 0) + b * ((xss !! 2) !! 1) + c * ((xss !! 2) !! 2)


--------------------------------------------------------------------------------
-- TESTING
--------------------------------------------------------------------------------

-- | Property to test vector addition and subtraction.
prop_addAndSubV :: Vector -> Vector -> Bool
prop_addAndSubV x y = (addV x y) == (subV x (negateV y))

-- | Property to test vector length functions.
prop_lengthV :: Vector -> Bool
prop_lengthV x = (lengthV x) == (sqrt (lengthSqrV x))

-- | Property to test vector negation.
prop_negateV :: Vector -> Bool
prop_negateV x = x == negateV (negateV x)

-- | Property to test normalization.
prop_normalizeV :: Vector -> Property
prop_normalizeV x = (x /= (0.0, 0.0, 0.0)) ==> lengthV (normalizeV x) == 1.0
