--------------------------------------------------------------------------------
-- | 
-- Module: Renderer
--
-- Contains structures and functions that allow to render a scene.
-- 
--------------------------------------------------------------------------------

module Renderer (
    -- * Functions
    render      -- :: Scene -> Camera -> (Int,Int) -> Bitmap
    ) where

import Bitmap
import BitmapTools
import Colors
import Material
import Primitives
import Ray
import Scene
import Utils


--------------------------------------------------------------------------------
-- SUPPLEMENTARY FUNCTIONS
--------------------------------------------------------------------------------

-- | Returns the primitive, which is being intersected by the ray first and the
-- type of that intersection. If there is no intersection (i.e. 'Miss'), then
-- the second element of the pair should be ignored.
--
-- Contains naive (linear) implementation of intersection search.
closestIntr :: Scene -> Ray -> (Intersection, Primitive)
closestIntr []     ray = (Miss, makeSphere (1,1,1) 1 (solidMaterial black))
closestIntr (p:ps) ray = if cDist < rDist then (curInt, p) else rest
    where
    curInt = intersects p ray   -- Intersection with current primitive.
    rest   = closestIntr ps ray -- Closest intersection from the remaining primitives.

    cDist = intDistance curInt      -- Distance to current intersection.
    rDist = intDistance (fst rest)  -- Distance to intersection from remaining primitives.

    -- Function, which returns the distance to the intersection.
    intDistance (Miss)      = 1e100
    intDistance (Outside x) = x
    intDistance (Inside x)  = x


-- | Returns the coordinates of the point, where a ray intersects some 
-- primitive.
intrPoint :: Ray -> Intersection -> Vector
intrPoint (orig, dir) (Miss)      = error "No intersection."
intrPoint (orig, dir) (Outside t) = addV orig (multScV t dir)
intrPoint (orig, dir) (Inside t)  = addV orig (multScV t dir)


-- | Returns the direction of a ray, reflected off some plane with specified
-- normal.
rayReflection :: Vector -> Vector -> Vector
rayReflection normal dir = subV dir (multScV (2 * dotProduct dir normal) normal)


--------------------------------------------------------------------------------
-- RENDERING FUNCTIONS
--------------------------------------------------------------------------------

-- | Traces a ray in the specified scene.
traceRay :: SceneOptions -> Scene -> Ray -> Int -> Vector
traceRay options scene ray depth 
    | scene == []       = skyboxCol
    | depth >= maxDepth = skyboxCol
    | otherwise         = case closestIntr scene ray of
                                (Miss, _)      -> skyboxCol
                                (Outside d, p) -> calcColor d p
                                (Inside  d, p) -> calcColor d p
    where
    -- Bisecting scene properties.
    ambLight  = ambientLight options
    maxDepth  = maxRayDepth options
    skyboxCol = skyboxColor options

    -- Function, which calculates the color of the ray at current intersection.
    calcColor dist prim = if isLight prim 
                            then getMaterialColor prim 
                          else if isPortal prim
                            then traceRay options scene (getPortalDest prim (makeRay intrPt dir)) (depth + 1)
                          else finalColor
        where
        -- Properties of primitive's material.
        primDiffuse  = getDiffuse prim
        primSpecular = getSpecular prim
        primReflect  = getReflection prim
        primShine    = getShininess prim
        -- Incoming ray.
        (orig, dir)  = ray

        -- Coordinates of ray's intersection point.
        intrPt = intrPoint ray (Outside dist)
        -- Primitive's normal at the intersection point.
        normal = getNormal prim intrPt

        -- Primitive's texture color at the intersection point.
        primColor    = multV (getMaterialColor prim) (getTextureColor prim intrPt)

        -- All light sources in the scene.
        lights    = filter isLight scene
        -- Directions of all light sources.
        lightDirs = [normalizeV (subV (getLightSource lt) intrPt) | lt <- lights]
        -- Shading of each light source.
        shades    = [calcShade lt lDir | (lt, lDir) <- zip lights lightDirs]

        -- Illumination of the given point.
        lightColor = illumination
        -- Color of the reflected ray.
        reflColor  = reflectedRayColor
        -- Final color of the ray.
        finalColor = addListV [lightColor, reflColor]

        -- Calculates shade of a light source.
        calcShade lt lDir = if clIntr == lt then 1.0 else 0.0
            where
            newRay      = makeRay (addV intrPt (multScV epsilon lDir)) lDir
            (_, clIntr) = closestIntr scene newRay

        -- Calculates the illumination of the current intersection point.
        illumination = finalCol
            where
            -- Ambient lighting.
            ambColor = multScV ambLight primColor

            -- Diffusion lighting value of each light source.
            diffLtValues = [(dotProduct normal lDir) * primDiffuse | lDir <- lightDirs]
            -- Color contribution of each light source to diffusion lighting.
            diffLtColors = [multV primColor (multScV (ltRating * shade) (getMaterialColor lt)) | (ltRating, shade, lt) <- zip3 diffLtValues shades lights, ltRating > 0]
            -- Diffusion lighting color component.
            diffColor    = addListV diffLtColors

            -- Reflections of rays, coming from all light sources.
            ltRefls = [rayReflection normal lDir | lDir <- lightDirs]
            -- Specular lighting value of each reflected ray.
            dots = [dotProduct dir refl | refl <- ltRefls]
            -- Color contribution of each light source to specular lighting.
            specLtColors = [multScV ((dot ^ primShine) * shade * primSpecular) (getMaterialColor lt) | (lt, shade, dot) <- zip3 lights shades dots, dot > 0]
            -- Specular lighting color component.
            specColor = addListV specLtColors

            -- Final color, provided by illumination.
            finalCol = addListV [ambColor, diffColor, specColor]

        -- Calculates the color of the reflected ray.
        reflectedRayColor = if primReflect > 0.0 then finalCol else (0,0,0)
            where
            -- Creating the direction of reflected ray.
            reflDir    = rayReflection normal dir
            -- Creating the reflected ray.
            reflRay    = makeRay (addV intrPt (multScV epsilon reflDir)) reflDir
            -- Traces the reflected ray and calculates its color
            reflColor  = traceRay options scene reflRay (depth + 1)
            -- Calculates the final color of reflected ray
            finalCol   = multV (getMaterialColor prim) (multScV primReflect reflColor)


-- | Renders the scene as a bitmap with specified dimensions.
render :: SceneOptions -> Camera -> Scene -> (Int,Int) -> Bitmap
render options camera scene (w,h) = result
    where
    -- Extracting camera data.
    camOrigin = cameraOrigin camera
    scrUL     = scrULCorner camera
    scrWidth  = scrWdVector camera
    scrHeight = scrHtVector camera
    -- Extracting scene data
    sSamples = fromIntegral (aaSamples options)

    -- Deltas for ray calculation
    deltaX = multScV (1 / (fromIntegral w)) scrWidth
    deltaY = multScV (1 / (fromIntegral h)) scrHeight

    -- Enumerating all pixels.
    pixels  = [ (x,y) | y <- [0..h-1], x <- [0..w-1] ]
    -- Calculating the colors of every pixel.
    colors  = [ getPixelColor (fromIntegral x) (fromIntegral y) | (x,y) <- pixels ]
    -- Combining results into a list of 'PixelChange'.
    changes = zip pixels colors
    -- Applying results to a bitmap.
    result  = setPixels (makeBitmap w h) changes

    -- Calculates the color of a specified pixel.
    getPixelColor x y = vector2rgb pixelColor
        where
        -- Partial rays of each pixel.
        partialRays = [getRay (x + a / sSamples) (y + b / sSamples) | a <- [1..sSamples], b <- [1..sSamples]]
        -- Accumulated color from each ray.
        accColor = addListV [traceRay options scene r 0 | r <- partialRays]
        -- Final color of the pixel.
        pixelColor = multScV (1 / sSamples^2) accColor

    -- Returns a ray, pointing from camera's original position to some point in
    -- screen plane.
    getRay x y = makeRay camOrigin rDir
        where
        -- Pixel's location on the screen rectangle.
        scrPt = addListV [scrUL, multScV y deltaY, multScV x deltaX]
        -- Direction of the ray.
        rDir  = normalizeV (subV scrPt camOrigin)

