--------------------------------------------------------------------------------
-- | 
-- Module: Scene
--
-- Contains structures and functions, which define a renderable scene.
-- 
--------------------------------------------------------------------------------

module Scene (
    -- * Types
    SceneOptions,
    Camera,
    Scene,

    -- * Constructor functions
    makeCamera,         -- :: Vector -> Vector -> Camera
    makeSceneOptions,   -- :: Double -> Int -> ColorRGB -> Int -> SceneOptions

    -- * Accessors
    aaSamples,      -- :: SceneOptions -> Int
    ambientLight,   -- :: SceneOptions -> Double
    maxRayDepth,    -- :: SceneOptions -> Int
    skyboxColor,    -- :: SceneOptions -> ColorVector

    cameraOrigin,   -- :: Camera -> Vector
    scrULCorner,    -- :: Camera -> Vector
    scrWdVector,    -- :: Camera -> Vector
    scrHtVector     -- :: Camera -> Vector

    ) where

import Colors
import Primitives
import Ray
import Utils


--------------------------------------------------------------------------------
-- SCENE PROPERTIES
--------------------------------------------------------------------------------

-- | Structure, which contains the properties of the scene.
data SceneOptions =
    ScOpts { ambientLight :: Double,        -- ^ Ambient light level of the scene.
             maxRayDepth  :: Int,           -- ^ Maximum depth or ray rendering.
             skyboxColor  :: ColorVector,   -- ^ Color of the skybox.
             aaSamples    :: Int            -- ^ Anti-aliasing samples.
    }

-- | Creates a correct SceneOptions object. The arguments, in order, are the
-- ambient level of the scene (a double between 0.0 and 1.0), maximum depth of
-- ray rendering and color of the skybox.
--
-- (There's gotta be an easier way to do this)
makeSceneOptions :: Double -> Int -> ColorRGB -> Int -> SceneOptions
makeSceneOptions ambLight maxDepth skyboxCol sSamples
    | (ambLight < 0.0) || (ambLight > 1.0) = error "Invalid ambient light level."
    | (maxDepth <= 0)                      = error "Invalid depth value."
    | (sSamples <= 0)                      = error "Invalid amount of AA samples."
    | otherwise = ScOpts ambLight maxDepth (rgb2vector skyboxCol) sSamples


--------------------------------------------------------------------------------
-- CAMERA
--------------------------------------------------------------------------------

-- | This type represents a camera, which is used to render the scene with.
-- Currently camera is assumed to be at location (0, 0, camPos).
data Camera =
    Cam { cameraOrigin :: Vector,   -- ^ Origin of the camera.
          scrULCorner  :: Vector,   -- ^ Location of the upper left corner of the screen rectangle.
          scrWdVector  :: Vector,   -- ^ Vector, defining the width of the screen rectangle.
          scrHtVector  :: Vector    -- ^ Vector, defining the height of the screen rectangle.
    }
    deriving(Eq, Read, Show)

-- | Creates a camera, which looks at some specified point from the origin. 
makeCamera :: Vector -> Vector -> Camera
makeCamera camOrig camTarget = newCamera
    where
    -- Fixed origin of the camera (camera's origin and target points must not
    -- have equal Z coordinates).
    camOrig' = let (x,y,z)  = camOrig
                   (_,_,z1) = camTarget
               in  if (z == z1) then (x, y, z + epsilon) else (x, y, z)

    -- Distance to the target.
    tgtDist = lengthV (subV camTarget camOrig')
    -- Offsets to calculate the screen rectangle.
    scrOffX = tgtDist * 0.8
    scrOffY = tgtDist * 0.8

    -- Axis, which define the camera.
    zAxis = normalizeV (subV camTarget camOrig')
    xAxis = normalizeV (crossVector (0, 1, 0) zAxis)
    yAxis = normalizeV (crossVector xAxis (multScV (-1) zAxis))

    -- Corner coordinates of the screen rectangle.
    newUL   = addListV [camTarget, multScV (-scrOffX) xAxis, multScV (scrOffY) yAxis]
    newLL   = addListV [camTarget, multScV (-scrOffX) xAxis, multScV (-scrOffY) yAxis]
    newUR   = addListV [camTarget, multScV (scrOffX) xAxis, multScV (scrOffY) yAxis]

    -- Vectors, representing the width and height of the screen rectangle.
    scrWdV = subV newUR newUL
    scrHtV = subV newLL newUL

    -- The new camera object.
    newCamera = Cam camOrig' newUL scrWdV scrHtV


--------------------------------------------------------------------------------
-- OTHER TYPES
--------------------------------------------------------------------------------

-- | A type, respresenting a scene of primitives.
type Scene = [Primitive]
